#<profile4xi profile = "02" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "joinablequeue4xi.py">
#<abstract4xi abstract = "provides class JoinableQueue"/>
#<refer4xi reference = "queues.py"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = [ 'JoinableQueue4Xi' ]
#"
#/>

# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
try:                 # system ones
    import os
    import sys
    import signal
except:
    pass
try:                 # threading ones       
    import threading                
    import _thread   
except:
    pass
try:
    import _winapi
except:
    pass

                
try:                 # most-common ones
    import time
    import weakref         
except:
    pass
try:                 # rare ones
    import collections
    import atexit
    from queue import Empty, Full          
except:
    pass 
#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
try:                             # Binary ones
    import _multiprocessing
except:
    pass
try:                             # Python ones
    from multiprocessing.queues import JoinableQueue
except:
    print("joinablequeue4xi.py: *** failed in from multiprocessing.queues import JoinableQueue ***")
#"
#/>



# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from ace_object4xi import ace_object4xi
from condition4xi import Condition4Xi
from queue4xi import Queue4Xi
from reason_object4xi import reason_object4xi
from semaphore4xi import Semaphore4Xi
from sense_object4xi import sense_object4xi
from timeout4xi import omega2timeout4xi
from print4xi import print4xi
from time4xi import time4xi    
#"
#/>




# =================================================================================================
#<class4xi class = "JoinableQueue4Xi" base = "Queue4Xi" child = "None"
#   synopsis = "A queue type supporting join() and task_done() methods"
#      usage = "
#   managers.py:
#       SyncManager.register('JoinableQueue', queue.Queue)
#   syncmanager4xi.py
#       SyncManager4Xi.register('JoinableQueue4Xi', queue.Queue)
#"
#>
#<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
# =================================================================================================
class JoinableQueue4Xi(Queue4Xi):
    '''
    JoinableQueue4Xi class:  A queue type which also supports join() and task_done() methods
    '''    
    # ---------------------------------------------------------------------------------------------
    # Public API of this class
    # ---------------------------------------------------------------------------------------------    
    #<expose4xi expose = "
    _exposed_ = ('put', 'join','task_done',
                 'sense', 'reason', 'ace')          
    #"
    #/>
                                     
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "init-of-JoinableQueue4Xi" usage = "initializing JoinableQueue4Xi">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def __init__(self, maxsize=0):
        '''
        initializing JoinableQueue4Xi
        '''              
                                         
        class2print = "JoinableQueue4Xi"
        method2print = "__init__"
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 1/5) stepping" + "; (Step 1/5)-stepping-by-init-of-JoinableQueue4Xi: " + "~~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------    
        message2print = "{(Step 1/5a) started" + "; init-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
                        
        message2print = "{(Step 1/5b) passed" + "; maxsize-passed-to-init-of-JoinableQueue4Xi: " + str(maxsize) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        
                                
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 2/5) stepping" + "; (Step 2/5)-stepping-by-init-of-JoinableQueue4Xi: " + "~ Queue4Xi-init calling ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------             
        
        message2print = "{(Step 2/5a) passing" + "; maxsize-passing-to-Queue4Xi-init-by-init-of-JoinableQueue4Xi: " + str(maxsize) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
                
        message2print = "{(Step 2/5a) borning" + "; joinablequeue2xi-of-JoinableQueue4Xi-borning: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  

        message2print = "{(Step 2/5a) inheriting" + "; Queue4Xi-inheriting-by-joinablequeue2xi-of-JoinableQueue4Xi: " + str(Queue4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        
        Queue4Xi.__init__(self, maxsize)
        
        message2print = "{(Step 2/5a) performed" + "; Queue4Xi-init-performed-by-init-of-JoinableQueue4Xi: " + str(Queue4Xi.__init__) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 3/5) stepping" + "; (Step 3/5)-stepping-by-init-of-JoinableQueue4Xi: " + "~ setting unfinished tasks ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------                        
        
        self._unfinished_tasks = Semaphore4Xi(0)
        
        message2print = "{(Step 3/5) set" + "; unfinished-tasks-of-joinablequeue2xi-set-by-init-of-JoinableQueue4Xi to: " + str(self._unfinished_tasks)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 4/5) stepping" + "; (Step 4/5)-stepping-by-init-of-JoinableQueue4Xi: " + "~ setting condition ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------                       
        
        self._cond = Condition4Xi()

        message2print = "{(Step 4/5) set" + "; condition-of-joinablequeue2xi-set-by-init-of-JoinableQueue4Xi to: " + str(self._cond)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 5/5) stepping" + "; (Step 5/5)-stepping-by-init-of-JoinableQueue4Xi: " + "~ finishing ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------    
                                 
        message2print = "{(Step 5/5a) inherited" + "; Queue4Xi-inherited-by-joinablequeue2xi-of-JoinableQueue4Xi: " + str(Queue4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                            
        message2print = "{(Step 5/5b) activated" + "; joinablequeue2xi-of-JoinableQueue4Xi-activated: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
                 
                                        
        message2print = "{(Step 5/5c) returning" + "; init-of-JoinableQueue4Xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
                        
                
                                 
                
    #</methodize4xi>
                
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "getstate-of-JoinableQueue4Xi" usage = "getting state">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def __getstate__(self):
        '''
        getting state
        '''              
        class2print = "JoinableQueue4Xi"
        method2print = "__getstate__"
                
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 1/2) stepping" + "; (Step 1/2)-stepping-by-getstate-of-JoinableQueue4Xi: " + "~~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------                            

        
        message2print = "{started" + "; getstate-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 2/2) stepping" + "; (Step 2/2)-stepping-by-getstate-of-JoinableQueue4Xi: " + "~ Queue4Xi-getstate-returning ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------                          
        message2print = "{(Step 2/2a) passing" + "; self-passing-to-Queue4Xi-by-getstate-of-JoinableQueue4Xi: " + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
        
        message2print = "{(Step 2/2b) passing" + "; condition-of-joinablequeue2xi-passing-to-Queue4Xi-by-getstate-of-JoinableQueue4Xi: " + str(self._cond) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
        
        message2print = "{(Step 2/2c) passing" + "; unfinished-tasks-of-joinablequeue2xi-passing-to-Queue4Xi-by-getstate-of-JoinableQueue4Xi: " + str(self._unfinished_tasks) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
                      
                          
        message2print = "{(Step 2/2d) returning" + "; Queue4Xi-getstate-returning-by-getstate-of-JoinableQueue4Xi: " + str(Queue4Xi.__getstate__(self) + (self._cond, self._unfinished_tasks)) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                             
        return Queue4Xi.__getstate__(self) + (self._cond, self._unfinished_tasks)
        
    #</methodize4xi>
            
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "setstate-of-JoinableQueue4Xi" usage = "setting state">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def __setstate__(self, state):
        '''
        setting state
        '''
        
        class2print = "JoinableQueue4Xi"
        method2print = "__setstate__"
                              
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 1/4) stepping" + "; (Step 1/4)-stepping-by-setstate-of-JoinableQueue4Xi: " + "~~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------

        message2print = "{(Step 1/4a) started" + "; setstate-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
                        
        message2print = "{(Step 1/4b) passed" + "; state-passed-to-setstate-of-JoinableQueue4Xi: " + str(state) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 2/4) stepping" + "; (Step 2/4)-stepping-by-setstate-of-JoinableQueue4Xi: " + "~ calling Queue4Xi-setstate ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------
                 
        message2print = "{(Step 2/4a) passing" + "; state-passing-to-Queue4Xi-setstate-by-setstate-of-JoinableQueue4Xi: " + str(state[:-2]) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)            
                
                                 
        Queue4Xi.__setstate__(self, state[:-2])
        
        message2print = "{performed" + "; Queue4Xi-setstate-performed-by-setstate-of-JoinableQueue4Xi: " + str(Queue4Xi.__setstate__) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
            
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 3/4) stepping" + "; (Step 3/4)-stepping-by-setstate-of-JoinableQueue4Xi: " + "~ setting condition and unfinished tasks ~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------
                     
        self._cond, self._unfinished_tasks = state[-2:]
        
        message2print = "{(Step 3/4a) set" + "; condition-of-joinablequeue2xi-set-by-setstate-of-JoinableQueue4Xi to: " + str(self._cond)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
                
        message2print = "{(Step 3/4b) set" + "; unfinished-tasks-of-joinablequeue2xi-set-by-setstate-of-JoinableQueue4Xi to: " + str(self._unfinished_tasks)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
                  
        # -------------------------------------------------------------------------------------------------------------------------------------------
        message2print = "{(Step 4/4) stepping" + "; (Step 1/4)-stepping-by-setstate-of-JoinableQueue4Xi: " + "~~"  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        # -------------------------------------------------------------------------------------------------------------------------------------------
                                
                
        message2print = "{(Step 4/4) returning" + "; setstate-of-JoinableQueue4Xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
                        
                
                                
    #</methodize4xi>
      
               
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "put-of-JoinableQueue4Xi" usage = "putting">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def put(self, item, block=True, timeout=None):
        '''
        putting
        ''' 
        class2print = "JoinableQueue4Xi"
        method2print = "put"                     
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 1/4)
        # -------------------------------------------------------------------------------------------------------------------------------------------                               

        
        message2print = "{(Step 1/4a) started" + "; put-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
  
                        
        message2print = "{(Step 1/4b) passed" + "; item-passed-to-put-of-JoinableQueue4Xi: " + str(item) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
                        
        message2print = "{(Step 1/4c) passed" + "; block-passed-to-put-of-JoinableQueue4Xi: " + str(block) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        
        timeout = omega2timeout4xi(class2print, method2print) 

        message2print = "{(Step 1/4d) set" + "; timeout-set-by-put-of-JoinableQueue4Xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)      
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 2/4)
        # -------------------------------------------------------------------------------------------------------------------------------------------                              
        assert not self._closed
        if not self._sem.acquire(block, timeout):
            message2print = "{(Step 2/4z) failed" + "; block-failed-by-put-of-JoinableQueue4Xi: " + str(block) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
            raise Full
            
        self._notempty.acquire()
        message2print = "{(Step 2/4a) performed" + "; notempty-acquire-performed-by-put-of-JoinableQueue4Xi: " + str(self._notempty.acquire) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #              
              
        self._cond.acquire()    
        message2print = "{(Step 2/4b) performed" + "; cond-acquire-performed-by-put-of-JoinableQueue4Xi: " + str(self._cond.acquire) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #     
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 3/4)
        # -------------------------------------------------------------------------------------------------------------------------------------------                        
        try:
            if self._thread is None:
                self._start_thread()
                message2print = "{(Step 3/4a) performed" + "; start-thread-performed-by-put-of-JoinableQueue4Xi: " + str(self._start_thread) + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                # 
                 
            self._buffer.append(obj)
            message2print = "{(Step 3/4b) performed" + "; append-buffer-performed-by-put-of-JoinableQueue4Xi: " + str(self._buffer.append) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
                
            self._unfinished_tasks.release()
            message2print = "{(Step 3/4c) performed" + "; release-unfinished-tasks-performed-by-put-of-JoinableQueue4Xi: " + str(self._unfinished_tasks.release) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
            
            self._notempty.notify()
            message2print = "{(Step 3/4d) performed" + "; notempty-notify-performed-by-put-of-JoinableQueue4Xi: " + str(self._notempty.notify) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
            
            
        finally:
            self._cond.release()
            message2print = "{(Step 3/4e) performed" + "; release-cond-performed-by-put-of-JoinableQueue4Xi: " + str(self._cond.release) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
            
            self._notempty.release()
            message2print = "{(Step 3/4f) performed" + "; release-notempty-performed-by-put-of-JoinableQueue4Xi: " + str(self._notempty.release) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            # 
            
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 4/4)
        # -------------------------------------------------------------------------------------------------------------------------------------------                          
        message2print = "{(Step 4/4) returning" + "; put-method-of-JoinableQueue4Xi returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
         
    #</methodize4xi>
             
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "task-done-of-JoinableQueue4Xi" usage = "task done">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def task_done(self):
        '''
        task done
        '''              
         
        class2print = "JoinableQueue4Xi"
        method2print = "task_done"
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 1/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                                
     
        message2print = "{(Step 1/5) started" + "; task-done-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 2/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------               
        self._cond.acquire()
        message2print = "{(Step 2/5) performed" + "; cond-acquire-performed-by-task-done-of-JoinableQueue4Xi: " + str(self._cond.acquire) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #    
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 3/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------             
           
        try:
            if not self._unfinished_tasks.acquire(False):
                
                message2print = "{(Step 3/5a) hit" + "; ValueError-hit-by-task-done-of-JoinableQueue4Xi: " + " *** JoinableQueue4Xi: task_done() called too many times *** " + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                                    
                raise ValueError('JoinableQueue4Xi: task_done() called too many times')
            if self._unfinished_tasks._semlock._is_zero():
                
                self._cond.notify_all()
                
                message2print = "{(Step 3/5b) performed" + "; cond-notify-all-performed-by-task-done-of-JoinableQueue4Xi: " + str(self._cond.notify_all) + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)             
                                
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 4/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                         
                     
        finally:
            self._cond.release()
            
            message2print = "{(Step 4/5) performed" + "; cond-release-performed-by-task-done-of-JoinableQueue4Xi: " + str(self._cond.release) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)             
                            
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 5/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                       
        message2print = "{(Step 5/5) returning" + "; task-done-of-JoinableQueue4Xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
                        
                
                                  
    #</methodize4xi>
                
                
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "join-of-JoinableQueue4Xi" usage = "joins queue">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------  
    def join(self):
        '''
        joins queue
        ''' 
         
        class2print = "JoinableQueue4Xi"
        method2print = "join"                       
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 1/5
        # -------------------------------------------------------------------------------------------------------------------------------------------                               

        message2print = "{started" + "; join-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
  
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 2/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                         
             
        self._cond.acquire()
        
        message2print = "{(Step 2/5) performed" + "; condition-acquire-of-joinablequeue2xi-performed-by-join-of-JoinableQueue4Xi: " + str(self._cond.acquire) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)             
                        
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 3/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                           
                
        try:
            if not self._unfinished_tasks._semlock._is_zero():
                self._cond.wait()
                
                message2print = "{(Step 3/5) performed" + "; condition-wait-of-joinablequeue2xi-performed-by-join-of-JoinableQueue4Xi: " + str(self._cond.wait) + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print) 
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 4/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                            
                                         
        finally:
            self._cond.release()
            
            message2print = "{(Step 4/5) performed" + "; condition-release-of-joinablequeue2xi-performed-by-join-of-JoinableQueue4Xi: " + str(self._cond.release) + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)             
                            
        
        # -------------------------------------------------------------------------------------------------------------------------------------------
        # (Step 5/5)
        # -------------------------------------------------------------------------------------------------------------------------------------------                              
        
        message2print = "{(Step 5/5) returning" + "; join-of-JoinableQueue4Xi-returning: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print) 
                        
                
                                 
    #</methodize4xi>
    
                                       
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "sense-of-JoinableQueue4Xi" usage = "senses JoinableQueue4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------
    def sense(self):
        '''
        senses JoinableQueue4Xi class objects
        '''
                                
        class2print = 'JoinableQueue4Xi'
        method2print = 'sense'
        
        # ---------------------------------------------------------------------------------------------
        #
        # ---------------------------------------------------------------------------------------------          
        message2print = "{started" + "; sense-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
  
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------  
        message2print = "{sensing" + "; JoinableQueue4Xi-sensing-by-sense-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                         
        
        # -----------------------------------------------------------------------------------------
        # get sensed knowledge for this object
        # -----------------------------------------------------------------------------------------        
        kb2xi = sense_object4xi(self)
        
        message2print = "{got" + "; kb2xi-got-by-sense-of-JoinableQueue4Xi: " + str(kb2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
        
        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # -----------------------------------------------------------------------------------------         
        try:
            kb2xi['condition-of-joinablequeue2xi'] = self._cond     
            kb2xi['unfinished-tasks-of-joinablequeue2xi'] = self._unfinished_tasks
        except:
            message2print = "{yielded" + "; kb2xi-yielded-by-sense-of-JoinableQueue4Xi: " + str(kb2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
        
              
        message2print = "{sensed" + "; JoinableQueue4Xi-sensed-by-sense-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)    
        
        # -----------------------------------------------------------------------------------------
        # 
        # -----------------------------------------------------------------------------------------         
        message2print = "{returning" + "; kb2xi-returning-by-sense-of-JoinableQueue4Xi: " + str(kb2xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                          
          
        return kb2xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        

    #</methodize4xi>
            
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "reason-of-JoinableQueue4Xi" usage = "reasons JoinableQueue4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------
    def reason(self):
        '''
        reasons JoinableQueue4Xi class objects
        '''
                             
        class2print = 'JoinableQueue4Xi'
        method2print = 'reason'

        # ---------------------------------------------------------------------------------------------
        #
        # ---------------------------------------------------------------------------------------------          
        message2print = "{started" + "; reason-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
        
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------
        message2print = "{reasoning" + "; JoinableQueue4Xi-reasoning-by-reason-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
                
        
        # -----------------------------------------------------------------------------------------
        # get reasoned knowledge for this object
        # -----------------------------------------------------------------------------------------         
        kb4xi = reason_object4xi(self)
        
        message2print = "{got" + "; kb4xi-got-by-reason-of-JoinableQueue4Xi: " + str(kb4xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                
        
        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # -----------------------------------------------------------------------------------------          
        try:
            kb4xi['unfinished-tasks-semlock-of-joinablequeue2xi'] = self._unfinished_tasks._semlock._is_zero()         
        except:
            message2print = "{yielded" + "; kb4xi-yielded-by-reason-of-JoinableQueue4Xi: " + str(kb4xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
       
        
        message2print = "{reasoned" + "; JoinableQueue4Xi-reasoned-by-reason-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        
        message2print = "{returning" + "; kb4xi-returning-by-reason-of-JoinableQueue4Xi: " + str(kb4xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                                 
          
        return kb4xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------            
    #</methodize4xi>
    
    
    # ---------------------------------------------------------------------------------------------
    #<methodize4xi method = "ace-of-JoinableQueue4Xi" usage = "aces JoinableQueue4Xi class objects">
    #<state4xi status = "ACK4review; ACK4tagging; NACK4moduletesting; NACK4systemtesting"/>
    # ---------------------------------------------------------------------------------------------
    def ace(self):
        '''
        aces JoinableQueue4Xi class objects 
        '''
          
        class2print = 'JoinableQueue4Xi'
        method2print = 'ace'
        # ---------------------------------------------------------------------------------------------
        #
        # ---------------------------------------------------------------------------------------------          
        
        message2print = "{started" + "; ace-of-JoinableQueue4Xi-started: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)   
       
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------
        message2print = "{acing" + "; JoinableQueue4Xi-acing-by-ace-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
                      
        
        # -----------------------------------------------------------------------------------------
        # get aced knowledge for this object
        # -----------------------------------------------------------------------------------------        
        kb8xi = ace_object4xi(self)  
        
        message2print = "{got" + "; kb8xi-got-by-ace-of-JoinableQueue4Xi: " + str(kb8xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
                


        # -----------------------------------------------------------------------------------------
        # add knowledge specific to this object
        # -----------------------------------------------------------------------------------------              
        try:
            kb8xi['release-of-joinablequeue2xi'] = str(self._unfinished_tasks.release)  
            kb8xi['acquire-of-joinablequeue2xi'] = str(self._unfinished_tasks.acquire)                
            kb8xi['semlock-check-of-joinablequeue2xi'] = str(self._unfinished_tasks._semlock._is_zero)  
            kb8xi['notifying-all-of-joinablequeue2xi'] = str(self._cond.notify_all) 
            kb8xi['waiting-of-joinablequeue2xi'] = str(self._cond.wait)  
            kb8xi['releasing-of-joinablequeue2xi'] = str(self._cond.release)                
        except:
            message2print = "{yielded" + "; kb8xi-yielded-by-ace-of-JoinableQueue4Xi: " + str(kb8xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print) 
        
                  
        message2print = "{aced" + "; JoinableQueue4Xi-aced-by-ace-of-JoinableQueue4Xi: " + str(JoinableQueue4Xi) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)         
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------        
        message2print = "{returning" + "; kb8xi-returning-by-ace-of-JoinableQueue4Xi: " + str(kb8xi) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
                             
        
        return kb8xi
        
        # -----------------------------------------------------------------------------------------
        #
        # -----------------------------------------------------------------------------------------            
    #</methodize4xi>
#</class4xi>
#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   test_none4xi.txt
#   sense-test-none4xi.xml
#   reason-test-none4xi.xml
#   ace-test-none4xi.xml
#"
#>


# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 
#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of joinablequeue4xi.p                             
# ------------------------------------------------------------------------------------------------- 
#"
#/>
